@@ -30,9 +30,7 @@
 
 #include <libavcodec/avcodec.h>
 #include <libavutil/stereo3d.h>
-
-#include <gst/gst.h>
-#include <gst/video/gstvideometa.h>
+#include <libavutil/opt.h>
 
 #include "gstav.h"
 #include "gstavcodecmap.h"
@@ -40,56 +38,16 @@
 #include "gstavvidenc.h"
 #include "gstavcfg.h"
 
-#define DEFAULT_VIDEO_BITRATE 300000    /* in bps */
-#define DEFAULT_VIDEO_GOP_SIZE 15
-
-#define DEFAULT_WIDTH 352
-#define DEFAULT_HEIGHT 288
-
-
-#define VIDEO_BUFFER_SIZE (1024*1024)
-
-enum
-{
-  /* FILL ME */
-  LAST_SIGNAL
-};
 
 enum
 {
   PROP_0,
-  PROP_BIT_RATE,
-  PROP_GOP_SIZE,
-  PROP_ME_METHOD,
-  PROP_BUFSIZE,
-  PROP_RTP_PAYLOAD_SIZE,
-  PROP_MAX_THREADS,
-  PROP_COMPLIANCE,
+  PROP_QUANTIZER,
+  PROP_PASS,
+  PROP_FILENAME,
   PROP_CFG_BASE,
 };
 
-#define GST_TYPE_ME_METHOD (gst_ffmpegvidenc_me_method_get_type())
-static GType
-gst_ffmpegvidenc_me_method_get_type (void)
-{
-  static GType ffmpegenc_me_method_type = 0;
-  static GEnumValue ffmpegenc_me_methods[] = {
-    {ME_ZERO, "None (Very low quality)", "zero"},
-    {ME_FULL, "Full (Slow, unmaintained)", "full"},
-    {ME_LOG, "Logarithmic (Low quality, unmaintained)", "logarithmic"},
-    {ME_PHODS, "phods (Low quality, unmaintained)", "phods"},
-    {ME_EPZS, "EPZS (Best quality, Fast)", "epzs"},
-    {ME_X1, "X1 (Experimental)", "x1"},
-    {0, NULL, NULL},
-  };
-  if (!ffmpegenc_me_method_type) {
-    ffmpegenc_me_method_type =
-        g_enum_register_static ("GstLibAVVidEncMeMethod", ffmpegenc_me_methods);
-  }
-  return ffmpegenc_me_method_type;
-}
-
-/* A number of function prototypes are given so we can refer to them later. */
 static void gst_ffmpegvidenc_class_init (GstFFMpegVidEncClass * klass);
 static void gst_ffmpegvidenc_base_init (GstFFMpegVidEncClass * klass);
 static void gst_ffmpegvidenc_init (GstFFMpegVidEnc * ffmpegenc);
@@ -116,7 +74,27 @@ static void gst_ffmpegvidenc_get_property (GObject * object,
 
 static GstElementClass *parent_class = NULL;
 
-/*static guint gst_ffmpegvidenc_signals[LAST_SIGNAL] = { 0 }; */
+#define GST_TYPE_FFMPEG_PASS (gst_ffmpeg_pass_get_type ())
+static GType
+gst_ffmpeg_pass_get_type (void)
+{
+  static GType ffmpeg_pass_type = 0;
+
+  if (!ffmpeg_pass_type) {
+    static const GEnumValue ffmpeg_passes[] = {
+      {0, "Constant Bitrate Encoding", "cbr"},
+      {AV_CODEC_FLAG_QSCALE, "Constant Quantizer", "quant"},
+      {AV_CODEC_FLAG_PASS1, "VBR Encoding - Pass 1", "pass1"},
+      {AV_CODEC_FLAG_PASS2, "VBR Encoding - Pass 2", "pass2"},
+      {0, NULL, NULL},
+    };
+
+    ffmpeg_pass_type =
+        g_enum_register_static ("GstLibAVEncPass", ffmpeg_passes);
+  }
+
+  return ffmpeg_pass_type;
+}
 
 static void
 gst_ffmpegvidenc_base_init (GstFFMpegVidEncClass * klass)
@@ -181,7 +159,6 @@ gst_ffmpegvidenc_class_init (GstFFMpegVidEncClass * klass)
 {
   GObjectClass *gobject_class;
   GstVideoEncoderClass *venc_class;
-  int caps;
 
   gobject_class = (GObjectClass *) klass;
   venc_class = (GstVideoEncoderClass *) klass;
@@ -191,46 +168,24 @@ gst_ffmpegvidenc_class_init (GstFFMpegVidEncClass * klass)
   gobject_class->set_property = gst_ffmpegvidenc_set_property;
   gobject_class->get_property = gst_ffmpegvidenc_get_property;
 
-  /* FIXME: could use -1 for a sensible per-codec default based on
-   * e.g. input resolution and framerate */
-  g_object_class_install_property (G_OBJECT_CLASS (klass), PROP_BIT_RATE,
-      g_param_spec_int ("bitrate", "Bit Rate",
-          "Target Video Bitrate", 0, G_MAXINT, DEFAULT_VIDEO_BITRATE,
-          G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS));
-  g_object_class_install_property (G_OBJECT_CLASS (klass), PROP_GOP_SIZE,
-      g_param_spec_int ("gop-size", "GOP Size",
-          "Number of frames within one GOP", 0, G_MAXINT,
-          DEFAULT_VIDEO_GOP_SIZE, G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS));
-  g_object_class_install_property (G_OBJECT_CLASS (klass), PROP_ME_METHOD,
-      g_param_spec_enum ("me-method", "ME Method", "Motion Estimation Method",
-          GST_TYPE_ME_METHOD, ME_EPZS,
-          G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS));
+  g_object_class_install_property (gobject_class, PROP_QUANTIZER,
+      g_param_spec_float ("quantizer", "Constant Quantizer",
+          "Constant Quantizer", 0, 30, 0.01f,
+          G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS | G_PARAM_CONSTRUCT));
 
-  g_object_class_install_property (G_OBJECT_CLASS (klass), PROP_BUFSIZE,
-      g_param_spec_int ("buffer-size", "Buffer Size",
-          "Size of the video buffers", 0, G_MAXINT, 0,
-          G_PARAM_READABLE | G_PARAM_STATIC_STRINGS));
-  g_object_class_install_property (G_OBJECT_CLASS (klass),
-      PROP_RTP_PAYLOAD_SIZE, g_param_spec_int ("rtp-payload-size",
-          "RTP Payload Size", "Target GOB length", 0, G_MAXINT, 0,
-          G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS));
+  g_object_class_install_property (gobject_class, PROP_PASS,
+      g_param_spec_enum ("pass", "Encoding pass/type",
+          "Encoding pass/type", GST_TYPE_FFMPEG_PASS, 0,
+          G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS | G_PARAM_CONSTRUCT));
 
-  caps = klass->in_plugin->capabilities;
-  if (caps & (CODEC_CAP_FRAME_THREADS | CODEC_CAP_SLICE_THREADS)) {
-    g_object_class_install_property (G_OBJECT_CLASS (klass), PROP_MAX_THREADS,
-        g_param_spec_int ("max-threads", "Maximum encode threads",
-            "Maximum number of worker threads to spawn. (0 = auto)",
-            0, G_MAXINT, 0, G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS));
-  }
-
-  g_object_class_install_property (G_OBJECT_CLASS (klass), PROP_COMPLIANCE,
-      g_param_spec_enum ("compliance", "Compliance",
-          "Adherence of the encoder to the specifications",
-          GST_TYPE_FFMPEG_COMPLIANCE, FFMPEG_DEFAULT_COMPLIANCE,
-          G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS));
+  g_object_class_install_property (gobject_class, PROP_FILENAME,
+      g_param_spec_string ("multipass-cache-file", "Multipass Cache File",
+          "Filename for multipass cache file", "stats.log",
+          G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS | G_PARAM_CONSTRUCT));
 
   /* register additional properties, possibly dependent on the exact CODEC */
-  gst_ffmpeg_cfg_install_property (klass, PROP_CFG_BASE);
+  gst_ffmpeg_cfg_install_properties (gobject_class, klass->in_plugin,
+      PROP_CFG_BASE, AV_OPT_FLAG_ENCODING_PARAM | AV_OPT_FLAG_VIDEO_PARAM);
 
   venc_class->start = gst_ffmpegvidenc_start;
   venc_class->stop = gst_ffmpegvidenc_stop;
@@ -251,26 +206,11 @@ gst_ffmpegvidenc_init (GstFFMpegVidEnc * ffmpegenc)
 
   GST_PAD_SET_ACCEPT_TEMPLATE (GST_VIDEO_ENCODER_SINK_PAD (ffmpegenc));
 
-  /* ffmpeg objects */
   ffmpegenc->context = avcodec_alloc_context3 (klass->in_plugin);
+  ffmpegenc->refcontext = avcodec_alloc_context3 (klass->in_plugin);
   ffmpegenc->picture = av_frame_alloc ();
   ffmpegenc->opened = FALSE;
-
   ffmpegenc->file = NULL;
-
-  ffmpegenc->bitrate = DEFAULT_VIDEO_BITRATE;
-  ffmpegenc->me_method = ME_EPZS;
-  ffmpegenc->buffer_size = 512 * 1024;
-  ffmpegenc->gop_size = DEFAULT_VIDEO_GOP_SIZE;
-  ffmpegenc->rtp_payload_size = 0;
-  ffmpegenc->compliance = FFMPEG_DEFAULT_COMPLIANCE;
-  ffmpegenc->max_threads = 0;
-
-  ffmpegenc->lmin = 2;
-  ffmpegenc->lmax = 31;
-  ffmpegenc->max_key_interval = 0;
-
-  gst_ffmpeg_cfg_set_defaults (ffmpegenc);
 }
 
 static void
@@ -278,14 +218,11 @@ gst_ffmpegvidenc_finalize (GObject * object)
 {
   GstFFMpegVidEnc *ffmpegenc = (GstFFMpegVidEnc *) object;
 
-  gst_ffmpeg_cfg_finalize (ffmpegenc);
-
   /* clean up remaining allocated data */
   av_frame_free (&ffmpegenc->picture);
   gst_ffmpeg_avcodec_close (ffmpegenc->context);
   av_free (ffmpegenc->context);
-
-  g_free (ffmpegenc->filename);
+  av_free (ffmpegenc->refcontext);
 
   G_OBJECT_CLASS (parent_class)->finalize (object);
 }
@@ -314,66 +251,25 @@ gst_ffmpegvidenc_set_format (GstVideoEncoder * encoder,
     }
   }
 
-  /* if we set it in _getcaps we should set it also in _link */
-  ffmpegenc->context->strict_std_compliance = ffmpegenc->compliance;
-
-  /* user defined properties */
-  ffmpegenc->context->bit_rate = ffmpegenc->bitrate;
-  ffmpegenc->context->bit_rate_tolerance = ffmpegenc->bitrate;
-  ffmpegenc->context->gop_size = ffmpegenc->gop_size;
-  ffmpegenc->context->me_method = ffmpegenc->me_method;
-  GST_DEBUG_OBJECT (ffmpegenc, "Setting avcontext to bitrate %d, gop_size %d",
-      ffmpegenc->bitrate, ffmpegenc->gop_size);
-
-  if (ffmpegenc->max_threads == 0) {
-    if (!(oclass->in_plugin->capabilities & CODEC_CAP_AUTO_THREADS))
-      ffmpegenc->context->thread_count = gst_ffmpeg_auto_max_threads ();
-    else
-      ffmpegenc->context->thread_count = 0;
-  } else
-    ffmpegenc->context->thread_count = ffmpegenc->max_threads;
-
-  /* RTP payload used for GOB production (for Asterisk) */
-  if (ffmpegenc->rtp_payload_size) {
-    ffmpegenc->context->rtp_payload_size = ffmpegenc->rtp_payload_size;
-  }
-
   /* additional avcodec settings */
-  /* first fill in the majority by copying over */
-  gst_ffmpeg_cfg_fill_context (ffmpegenc, ffmpegenc->context);
-
-  /* then handle some special cases */
-  ffmpegenc->context->lmin = (ffmpegenc->lmin * FF_QP2LAMBDA + 0.5);
-  ffmpegenc->context->lmax = (ffmpegenc->lmax * FF_QP2LAMBDA + 0.5);
-
-  if (ffmpegenc->interlaced) {
-    ffmpegenc->context->flags |=
-        CODEC_FLAG_INTERLACED_DCT | CODEC_FLAG_INTERLACED_ME;
-  }
-
-  /* some other defaults */
-  ffmpegenc->context->rc_strategy = 2;
-  ffmpegenc->context->b_frame_strategy = 0;
-  ffmpegenc->context->coder_type = 0;
-  ffmpegenc->context->context_model = 0;
-  ffmpegenc->context->scenechange_threshold = 0;
+  gst_ffmpeg_cfg_fill_context (G_OBJECT (ffmpegenc), ffmpegenc->context);
 
   /* and last but not least the pass; CBR, 2-pass, etc */
   ffmpegenc->context->flags |= ffmpegenc->pass;
   switch (ffmpegenc->pass) {
       /* some additional action depends on type of pass */
-    case CODEC_FLAG_QSCALE:
+    case AV_CODEC_FLAG_QSCALE:
       ffmpegenc->context->global_quality
           = ffmpegenc->picture->quality = FF_QP2LAMBDA * ffmpegenc->quantizer;
       break;
-    case CODEC_FLAG_PASS1:     /* need to prepare a stats file */
+    case AV_CODEC_FLAG_PASS1:  /* need to prepare a stats file */
       /* we don't close when changing caps, fingers crossed */
       if (!ffmpegenc->file)
         ffmpegenc->file = g_fopen (ffmpegenc->filename, "w");
       if (!ffmpegenc->file)
         goto open_file_err;
       break;
-    case CODEC_FLAG_PASS2:
+    case AV_CODEC_FLAG_PASS2:
     {                           /* need to read the whole stats file ! */
       gsize size;
 
@@ -412,18 +308,6 @@ gst_ffmpegvidenc_set_format (GstVideoEncoder * encoder,
 
   pix_fmt = ffmpegenc->context->pix_fmt;
 
-  /* max-key-interval may need the framerate set above */
-  if (ffmpegenc->max_key_interval) {
-    AVCodecContext *ctx;
-
-    /* override gop-size */
-    ctx = ffmpegenc->context;
-    ctx->gop_size = (ffmpegenc->max_key_interval < 0) ?
-        (-ffmpegenc->max_key_interval
-        * (ctx->time_base.den * ctx->ticks_per_frame / ctx->time_base.num))
-        : ffmpegenc->max_key_interval;
-  }
-
   /* some codecs support more than one format, first auto-choose one */
   GST_DEBUG_OBJECT (ffmpegenc, "picking an output format ...");
   allowed_caps = gst_pad_get_allowed_caps (GST_VIDEO_ENCODER_SRC_PAD (encoder));
@@ -633,27 +517,53 @@ stereo_gst_to_av (GstVideoMultiviewMode mview_mode)
   return AV_STEREO3D_2D;
 }
 
+static void
+gst_ffmpegvidenc_add_cc (GstBuffer * buffer, AVFrame * picture)
+{
+  GstVideoCaptionMeta *cc_meta;
+  gpointer iter = NULL;
+
+  while ((cc_meta =
+          (GstVideoCaptionMeta *) gst_buffer_iterate_meta_filtered (buffer,
+              &iter, GST_VIDEO_CAPTION_META_API_TYPE))) {
+    AVFrameSideData *sd;
+
+    if (cc_meta->caption_type != GST_VIDEO_CAPTION_TYPE_CEA708_RAW)
+      continue;
+
+    sd = av_frame_new_side_data (picture, AV_FRAME_DATA_A53_CC, cc_meta->size);
+    memcpy (sd->data, cc_meta->data, cc_meta->size);
+  }
+}
+
 static GstFlowReturn
-gst_ffmpegvidenc_handle_frame (GstVideoEncoder * encoder,
+gst_ffmpegvidenc_send_frame (GstFFMpegVidEnc * ffmpegenc,
     GstVideoCodecFrame * frame)
 {
-  GstFFMpegVidEnc *ffmpegenc = (GstFFMpegVidEnc *) encoder;
-  GstBuffer *outbuf;
-  gint ret = 0, c;
   GstVideoInfo *info = &ffmpegenc->input_state->info;
-  AVPacket *pkt;
-  int have_data = 0;
   BufferInfo *buffer_info;
+  guint c;
+  gint res;
+  GstFlowReturn ret = GST_FLOW_ERROR;
+  AVFrame *picture = NULL;
 
-  if (ffmpegenc->interlaced) {
-    ffmpegenc->picture->interlaced_frame = TRUE;
+  if (!frame)
+    goto send_frame;
+
+  picture = ffmpegenc->picture;
+
+  gst_ffmpegvidenc_add_cc (frame->input_buffer, picture);
+
+  if (ffmpegenc->context->flags & (AV_CODEC_FLAG_INTERLACED_DCT |
+          AV_CODEC_FLAG_INTERLACED_ME)) {
+    picture->interlaced_frame = TRUE;
     /* if this is not the case, a filter element should be used to swap fields */
-    ffmpegenc->picture->top_field_first =
+    picture->top_field_first =
         GST_BUFFER_FLAG_IS_SET (frame->input_buffer, GST_VIDEO_BUFFER_FLAG_TFF);
   }
 
   if (GST_VIDEO_INFO_MULTIVIEW_MODE (info) != GST_VIDEO_MULTIVIEW_MODE_NONE) {
-    AVStereo3D *stereo = av_stereo3d_create_side_data (ffmpegenc->picture);
+    AVStereo3D *stereo = av_stereo3d_create_side_data (picture);
     stereo->type = stereo_gst_to_av (GST_VIDEO_INFO_MULTIVIEW_MODE (info));
 
     if (GST_VIDEO_INFO_MULTIVIEW_FLAGS (info) &
@@ -663,64 +573,86 @@ gst_ffmpegvidenc_handle_frame (GstVideoEncoder * encoder,
   }
 
   if (GST_VIDEO_CODEC_FRAME_IS_FORCE_KEYFRAME (frame))
-    ffmpegenc->picture->pict_type = AV_PICTURE_TYPE_I;
+    picture->pict_type = AV_PICTURE_TYPE_I;
 
   buffer_info = g_slice_new0 (BufferInfo);
   buffer_info->buffer = gst_buffer_ref (frame->input_buffer);
 
   if (!gst_video_frame_map (&buffer_info->vframe, info, frame->input_buffer,
           GST_MAP_READ)) {
-    GST_ERROR_OBJECT (encoder, "Failed to map input buffer");
+    GST_ERROR_OBJECT (ffmpegenc, "Failed to map input buffer");
     gst_buffer_unref (buffer_info->buffer);
     g_slice_free (BufferInfo, buffer_info);
     gst_video_codec_frame_unref (frame);
-    return GST_FLOW_ERROR;
+    goto done;
   }
 
   /* Fill avpicture */
-  ffmpegenc->picture->buf[0] =
+  picture->buf[0] =
       av_buffer_create (NULL, 0, buffer_info_free, buffer_info, 0);
   for (c = 0; c < AV_NUM_DATA_POINTERS; c++) {
     if (c < GST_VIDEO_INFO_N_COMPONENTS (info)) {
-      ffmpegenc->picture->data[c] =
-          GST_VIDEO_FRAME_PLANE_DATA (&buffer_info->vframe, c);
-      ffmpegenc->picture->linesize[c] =
+      picture->data[c] = GST_VIDEO_FRAME_PLANE_DATA (&buffer_info->vframe, c);
+      picture->linesize[c] =
           GST_VIDEO_FRAME_COMP_STRIDE (&buffer_info->vframe, c);
     } else {
-      ffmpegenc->picture->data[c] = NULL;
-      ffmpegenc->picture->linesize[c] = 0;
+      picture->data[c] = NULL;
+      picture->linesize[c] = 0;
     }
   }
 
-  ffmpegenc->picture->format = ffmpegenc->context->pix_fmt;
-  ffmpegenc->picture->width = GST_VIDEO_FRAME_WIDTH (&buffer_info->vframe);
-  ffmpegenc->picture->height = GST_VIDEO_FRAME_HEIGHT (&buffer_info->vframe);
+  picture->format = ffmpegenc->context->pix_fmt;
+  picture->width = GST_VIDEO_FRAME_WIDTH (&buffer_info->vframe);
+  picture->height = GST_VIDEO_FRAME_HEIGHT (&buffer_info->vframe);
 
-  ffmpegenc->picture->pts =
+  picture->pts =
       gst_ffmpeg_time_gst_to_ff (frame->pts /
       ffmpegenc->context->ticks_per_frame, ffmpegenc->context->time_base);
 
-  have_data = 0;
+send_frame:
+  res = avcodec_send_frame (ffmpegenc->context, picture);
+
+  if (picture)
+    av_frame_unref (picture);
+
+  if (res == 0)
+    ret = GST_FLOW_OK;
+  else if (res == AVERROR_EOF)
+    ret = GST_FLOW_EOS;
+
+done:
+  return ret;
+}
+
+static GstFlowReturn
+gst_ffmpegvidenc_receive_packet (GstFFMpegVidEnc * ffmpegenc,
+    gboolean * got_packet, gboolean send)
+{
+  AVPacket *pkt;
+  GstBuffer *outbuf;
+  GstVideoCodecFrame *frame;
+  gint res;
+  GstFlowReturn ret = GST_FLOW_OK;
+
+  *got_packet = FALSE;
+
   pkt = g_slice_new0 (AVPacket);
 
-  ret =
-      avcodec_encode_video2 (ffmpegenc->context, pkt, ffmpegenc->picture,
-      &have_data);
+  res = avcodec_receive_packet (ffmpegenc->context, pkt);
 
-  av_frame_unref (ffmpegenc->picture);
-
-  if (ret < 0 || !have_data)
+  if (res == AVERROR (EAGAIN)) {
     g_slice_free (AVPacket, pkt);
-
-  if (ret < 0)
-    goto encode_fail;
-
-  /* Encoder needs more data */
-  if (!have_data) {
-    gst_video_codec_frame_unref (frame);
-    return GST_FLOW_OK;
+    goto done;
+  } else if (res == AVERROR_EOF) {
+    ret = GST_FLOW_EOS;
+    goto done;
+  } else if (res < 0) {
+    res = GST_FLOW_ERROR;
+    goto done;
   }
 
+  *got_packet = TRUE;
+
   /* save stats info if there is some as well as a stats file */
   if (ffmpegenc->file && ffmpegenc->context->stats_out)
     if (fprintf (ffmpegenc->file, "%s", ffmpegenc->context->stats_out) < 0)
@@ -728,24 +660,52 @@ gst_ffmpegvidenc_handle_frame (GstVideoEncoder * encoder,
           (("Could not write to file \"%s\"."), ffmpegenc->filename),
           GST_ERROR_SYSTEM);
 
+  /* Get oldest frame */
+  frame = gst_video_encoder_get_oldest_frame (GST_VIDEO_ENCODER (ffmpegenc));
+
+  if (send) {
+    outbuf =
+        gst_buffer_new_wrapped_full (GST_MEMORY_FLAG_READONLY, pkt->data,
+        pkt->size, 0, pkt->size, pkt, gst_ffmpegvidenc_free_avpacket);
+    frame->output_buffer = outbuf;
+
+    if (pkt->flags & AV_PKT_FLAG_KEY)
+      GST_VIDEO_CODEC_FRAME_SET_SYNC_POINT (frame);
+    else
+      GST_VIDEO_CODEC_FRAME_UNSET_SYNC_POINT (frame);
+  }
+
+  ret = gst_video_encoder_finish_frame (GST_VIDEO_ENCODER (ffmpegenc), frame);
+
+done:
+  return ret;
+}
+
+static GstFlowReturn
+gst_ffmpegvidenc_handle_frame (GstVideoEncoder * encoder,
+    GstVideoCodecFrame * frame)
+{
+  GstFFMpegVidEnc *ffmpegenc = (GstFFMpegVidEnc *) encoder;
+  GstFlowReturn ret;
+  gboolean got_packet;
+
+  ret = gst_ffmpegvidenc_send_frame (ffmpegenc, frame);
+
+  if (ret != GST_FLOW_OK)
+    goto encode_fail;
+
   gst_video_codec_frame_unref (frame);
 
-  /* Get oldest frame */
-  frame = gst_video_encoder_get_oldest_frame (encoder);
+  do {
+    ret = gst_ffmpegvidenc_receive_packet (ffmpegenc, &got_packet, TRUE);
+    if (ret != GST_FLOW_OK)
+      break;
+  } while (got_packet);
 
-  outbuf =
-      gst_buffer_new_wrapped_full (GST_MEMORY_FLAG_READONLY, pkt->data,
-      pkt->size, 0, pkt->size, pkt, gst_ffmpegvidenc_free_avpacket);
-  frame->output_buffer = outbuf;
+done:
+  return ret;
 
-  if (pkt->flags & AV_PKT_FLAG_KEY)
-    GST_VIDEO_CODEC_FRAME_SET_SYNC_POINT (frame);
-  else
-    GST_VIDEO_CODEC_FRAME_UNSET_SYNC_POINT (frame);
-
-  return gst_video_encoder_finish_frame (encoder, frame);
-
-  /* ERRORS */
+  /* We choose to be error-resilient */
 encode_fail:
   {
 #ifndef GST_DISABLE_GST_DEBUG
@@ -755,19 +715,16 @@ encode_fail:
         "avenc_%s: failed to encode buffer", oclass->in_plugin->name);
 #endif /* GST_DISABLE_GST_DEBUG */
     /* avoid frame (and ts etc) piling up */
-    return gst_video_encoder_finish_frame (encoder, frame);
+    ret = gst_video_encoder_finish_frame (encoder, frame);
+    goto done;
   }
 }
 
 static GstFlowReturn
 gst_ffmpegvidenc_flush_buffers (GstFFMpegVidEnc * ffmpegenc, gboolean send)
 {
-  GstVideoCodecFrame *frame;
-  GstFlowReturn flow_ret = GST_FLOW_OK;
-  GstBuffer *outbuf;
-  gint ret;
-  AVPacket *pkt;
-  int have_data = 0;
+  GstFlowReturn ret = GST_FLOW_OK;
+  gboolean got_packet;
 
   GST_DEBUG_OBJECT (ffmpegenc, "flushing buffers with sending %d", send);
 
@@ -775,64 +732,27 @@ gst_ffmpegvidenc_flush_buffers (GstFFMpegVidEnc * ffmpegenc, gboolean send)
   if (!ffmpegenc->opened)
     goto done;
 
-  while ((frame =
-          gst_video_encoder_get_oldest_frame (GST_VIDEO_ENCODER (ffmpegenc)))) {
-    pkt = g_slice_new0 (AVPacket);
-    have_data = 0;
+  ret = gst_ffmpegvidenc_send_frame (ffmpegenc, NULL);
 
-    ret = avcodec_encode_video2 (ffmpegenc->context, pkt, NULL, &have_data);
+  if (ret != GST_FLOW_OK)
+    goto done;
 
-    if (ret < 0) {              /* there should be something, notify and give up */
-#ifndef GST_DISABLE_GST_DEBUG
-      GstFFMpegVidEncClass *oclass =
-          (GstFFMpegVidEncClass *) (G_OBJECT_GET_CLASS (ffmpegenc));
-      GST_WARNING_OBJECT (ffmpegenc,
-          "avenc_%s: failed to flush buffer", oclass->in_plugin->name);
-#endif /* GST_DISABLE_GST_DEBUG */
-      g_slice_free (AVPacket, pkt);
-      gst_video_codec_frame_unref (frame);
+  do {
+    ret = gst_ffmpegvidenc_receive_packet (ffmpegenc, &got_packet, send);
+    if (ret != GST_FLOW_OK)
       break;
-    }
-
-    /* save stats info if there is some as well as a stats file */
-    if (ffmpegenc->file && ffmpegenc->context->stats_out)
-      if (fprintf (ffmpegenc->file, "%s", ffmpegenc->context->stats_out) < 0)
-        GST_ELEMENT_ERROR (ffmpegenc, RESOURCE, WRITE,
-            (("Could not write to file \"%s\"."), ffmpegenc->filename),
-            GST_ERROR_SYSTEM);
-
-    if (send && have_data) {
-      outbuf =
-          gst_buffer_new_wrapped_full (GST_MEMORY_FLAG_READONLY, pkt->data,
-          pkt->size, 0, pkt->size, pkt, gst_ffmpegvidenc_free_avpacket);
-      frame->output_buffer = outbuf;
-
-      if (pkt->flags & AV_PKT_FLAG_KEY)
-        GST_VIDEO_CODEC_FRAME_SET_SYNC_POINT (frame);
-      else
-        GST_VIDEO_CODEC_FRAME_UNSET_SYNC_POINT (frame);
-
-      flow_ret =
-          gst_video_encoder_finish_frame (GST_VIDEO_ENCODER (ffmpegenc), frame);
-    } else {
-      /* no frame attached, so will be skipped and removed from frame list */
-      gst_video_encoder_finish_frame (GST_VIDEO_ENCODER (ffmpegenc), frame);
-    }
-  }
+  } while (got_packet);
 
 done:
-
-  return flow_ret;
+  return ret;
 }
 
-
 static void
 gst_ffmpegvidenc_set_property (GObject * object,
     guint prop_id, const GValue * value, GParamSpec * pspec)
 {
   GstFFMpegVidEnc *ffmpegenc;
 
-  /* Get a pointer of the right type. */
   ffmpegenc = (GstFFMpegVidEnc *) (object);
 
   if (ffmpegenc->opened) {
@@ -841,69 +761,44 @@ gst_ffmpegvidenc_set_property (GObject * object,
     return;
   }
 
-  /* Check the argument id to see which argument we're setting. */
   switch (prop_id) {
-    case PROP_BIT_RATE:
-      ffmpegenc->bitrate = g_value_get_int (value);
+    case PROP_QUANTIZER:
+      ffmpegenc->quantizer = g_value_get_float (value);
       break;
-    case PROP_GOP_SIZE:
-      ffmpegenc->gop_size = g_value_get_int (value);
+    case PROP_PASS:
+      ffmpegenc->pass = g_value_get_enum (value);
       break;
-    case PROP_ME_METHOD:
-      ffmpegenc->me_method = g_value_get_enum (value);
-      break;
-    case PROP_BUFSIZE:
-      break;
-    case PROP_RTP_PAYLOAD_SIZE:
-      ffmpegenc->rtp_payload_size = g_value_get_int (value);
-      break;
-    case PROP_COMPLIANCE:
-      ffmpegenc->compliance = g_value_get_enum (value);
-      break;
-    case PROP_MAX_THREADS:
-      ffmpegenc->max_threads = g_value_get_int (value);
+    case PROP_FILENAME:
+      g_free (ffmpegenc->filename);
+      ffmpegenc->filename = g_value_dup_string (value);
       break;
     default:
-      if (!gst_ffmpeg_cfg_set_property (object, value, pspec))
+      if (!gst_ffmpeg_cfg_set_property (ffmpegenc->refcontext, value, pspec))
         G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
       break;
   }
 }
 
-/* The set function is simply the inverse of the get fuction. */
 static void
 gst_ffmpegvidenc_get_property (GObject * object,
     guint prop_id, GValue * value, GParamSpec * pspec)
 {
   GstFFMpegVidEnc *ffmpegenc;
 
-  /* It's not null if we got it, but it might not be ours */
   ffmpegenc = (GstFFMpegVidEnc *) (object);
 
   switch (prop_id) {
-    case PROP_BIT_RATE:
-      g_value_set_int (value, ffmpegenc->bitrate);
+    case PROP_QUANTIZER:
+      g_value_set_float (value, ffmpegenc->quantizer);
       break;
-    case PROP_GOP_SIZE:
-      g_value_set_int (value, ffmpegenc->gop_size);
+    case PROP_PASS:
+      g_value_set_enum (value, ffmpegenc->pass);
       break;
-    case PROP_ME_METHOD:
-      g_value_set_enum (value, ffmpegenc->me_method);
-      break;
-    case PROP_BUFSIZE:
-      g_value_set_int (value, ffmpegenc->buffer_size);
-      break;
-    case PROP_RTP_PAYLOAD_SIZE:
-      g_value_set_int (value, ffmpegenc->rtp_payload_size);
-      break;
-    case PROP_COMPLIANCE:
-      g_value_set_enum (value, ffmpegenc->compliance);
-      break;
-    case PROP_MAX_THREADS:
-      g_value_set_int (value, ffmpegenc->max_threads);
+    case PROP_FILENAME:
+      g_value_take_string (value, g_strdup (ffmpegenc->filename));
       break;
     default:
-      if (!gst_ffmpeg_cfg_get_property (object, value, pspec))
+      if (!gst_ffmpeg_cfg_get_property (ffmpegenc->refcontext, value, pspec))
         G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
       break;
   }
@@ -946,10 +841,6 @@ gst_ffmpegvidenc_stop (GstVideoEncoder * encoder)
   gst_ffmpeg_avcodec_close (ffmpegenc->context);
   ffmpegenc->opened = FALSE;
 
-  if (ffmpegenc->file) {
-    fclose (ffmpegenc->file);
-    ffmpegenc->file = NULL;
-  }
   if (ffmpegenc->input_state) {
     gst_video_codec_state_unref (ffmpegenc->input_state);
     ffmpegenc->input_state = NULL;
@@ -982,20 +873,16 @@ gst_ffmpegvidenc_register (GstPlugin * plugin)
   };
   GType type;
   AVCodec *in_plugin;
-
+  void *i = 0;
 
   GST_LOG ("Registering encoders");
 
-  /* build global ffmpeg param/property info */
-  gst_ffmpeg_cfg_init ();
-
-  in_plugin = av_codec_next (NULL);
-  while (in_plugin) {
+  while ((in_plugin = (AVCodec *) av_codec_iterate (&i))) {
     gchar *type_name;
 
     /* Skip non-AV codecs */
     if (in_plugin->type != AVMEDIA_TYPE_VIDEO)
-      goto next;
+      continue;
 
     /* no quasi codecs, please */
     if (in_plugin->id == AV_CODEC_ID_RAWVIDEO ||
@@ -1014,7 +901,7 @@ gst_ffmpegvidenc_register (GstPlugin * plugin)
         || in_plugin->id == AV_CODEC_ID_WRAPPED_AVFRAME
 #endif
         || in_plugin->id == AV_CODEC_ID_ZLIB) {
-      goto next;
+      continue;
     }
 
     /* No encoders depending on external libraries (we don't build them, but
@@ -1024,41 +911,41 @@ gst_ffmpegvidenc_register (GstPlugin * plugin)
       GST_DEBUG
           ("Not using external library encoder %s. Use the gstreamer-native ones instead.",
           in_plugin->name);
-      goto next;
+      continue;
     }
 
     if (strstr (in_plugin->name, "vaapi")) {
       GST_DEBUG
           ("Ignoring VAAPI encoder %s. We can't handle this outside of ffmpeg",
           in_plugin->name);
-      goto next;
+      continue;
     }
 
     if (strstr (in_plugin->name, "nvenc")) {
       GST_DEBUG
           ("Ignoring nvenc encoder %s. We can't handle this outside of ffmpeg",
           in_plugin->name);
-      goto next;
+      continue;
     }
 
     if (g_str_has_suffix (in_plugin->name, "_qsv")) {
       GST_DEBUG
           ("Ignoring qsv encoder %s. We can't handle this outside of ffmpeg",
           in_plugin->name);
-      goto next;
+      continue;
     }
 
     if (g_str_has_suffix (in_plugin->name, "_v4l2m2m")) {
       GST_DEBUG
           ("Ignoring V4L2 mem-to-mem encoder %s. We can't handle this outside of ffmpeg",
           in_plugin->name);
-      goto next;
+      continue;
     }
 
     /* only video encoders */
     if (!av_codec_is_encoder (in_plugin)
         || in_plugin->type != AVMEDIA_TYPE_VIDEO)
-      goto next;
+      continue;
 
     /* FIXME : We should have a method to know cheaply whether we have a mapping
      * for the given plugin or not */
@@ -1068,7 +955,7 @@ gst_ffmpegvidenc_register (GstPlugin * plugin)
     /* no codecs for which we're GUARANTEED to have better alternatives */
     if (!strcmp (in_plugin->name, "gif")) {
       GST_LOG ("Ignoring encoder %s", in_plugin->name);
-      goto next;
+      continue;
     }
 
     /* construct the type */
@@ -1100,9 +987,6 @@ gst_ffmpegvidenc_register (GstPlugin * plugin)
     }
 
     g_free (type_name);
-
-  next:
-    in_plugin = av_codec_next (in_plugin);
   }
 
   GST_LOG ("Finished registering encoders");
